#######################################################################
### MP Fénelon Sainte-Marie --- algorithme knn
#######################################################################

import numpy.random as rd
import numpy as np
import matplotlib.pyplot as plt
plt.clf()

#######################################################################

# Construction de N données aléatoires pour chacune des trois classes

N = 200
NbDeClasses = 3
Donnees = []
NbDeDonnees = N * NbDeClasses
delta = 0.5 # pour éloigner les trois familles de données les unes des autres

for i in range(N):
    Donnees.append([rd.random(),rd.random(),0])
for i in range(N):
    Donnees.append([rd.random()+delta,rd.random()+delta,1])
for i in range(N):
    Donnees.append([rd.random()+2*delta,rd.random()+2*delta,2])

# Mélange des données

rd.shuffle(Donnees)

# Affichage des données

for donnee in Donnees:
    if donnee[2]==0:
        plt.plot([donnee[0]],[donnee[1]],'ro')
    if donnee[2]==1:
        plt.plot([donnee[0]],[donnee[1]],'bo')
    if donnee[2]==2:
        plt.plot([donnee[0]],[donnee[1]],'go')
Couleurs = ['rouge','bleu','vert']

# 75% des données pour l'apprentissage et 25% pour le test

DonneesEntrainement = Donnees[0:int(NbDeDonnees*0.75)]
DonneesTest = Donnees[int(NbDeDonnees*0.75):]

# Distance euclidienne entre deux données

def distance2(d1,d2):
    # la distance euclidienne entre d1=[x1,y1,classe1] et d2=[x2,y2,classe2]
    resu = (d1[0]-d2[0])**2 + (d1[1]-d2[1])**2
    return np.sqrt(resu)

#######################################################################
# Prédiction de la classe d'une donnée à l'aide du jeu de données 'Entrainement'

def knnPrediction(Entrainement,donnee,k):
    # Prédiction de la classe (0, 1 ou 2) de l'élément 'donnee'
    # à l'aide des données d'entraînement, pour la méthode knn avec k voisins
    
    # Construction de la liste des couples [classe de d_i, distance de d_i à donnee]
    L = [] # La liste en question
    
    for data in Entrainement:
        dist = distance2(data,donnee)
        L.append([data[2],dist])
    
    # Tri de Distances selon la première coordonnée décroissante
    L.sort(key = lambda elem: elem[1])
    
    # Récupération des k plus proches voisins de donnee
    kPlusProches = L[:k]
    
    # Comptage des occurrences des trois classes des k plus proches voisins
    compteur = [0,0,0]
    for voisin in kPlusProches:
        compteur[voisin[0]] = compteur[voisin[0]] + 1
    
    # Parmi les k plus proches voisins, recherche de la classe de compteur le plus élevé
    maximum = 0
    for i in range(3):
        if compteur[i] > maximum:
            classe = i
            maximum = compteur[i]
    return classe

"""
DonneeTest = [1.1,0.9,7]
plt.plot([DonneeTest[0]],[DonneeTest[1]],'ko')
ClasseTest = knnPrediction(DonneesEntrainement,DonneeTest,5)
print('L\'échantillon',DonneeTest,' est de type',Couleurs[ClasseTest])
plt.axis([0,1+2*delta,0,1+2*delta])
"""

#######################################################################
# Prédiction de la classe des éléments de la liste 'Test' grâce à la liste 'Entrainement'

def knnTest(Entrainement,Test,k):
    # Prédiction de la classe (0, 1 ou 2) de tous les éléments de la liste Test
    # à l'aide des données de la liste Entraînement, pour la méthode knn avec k voisins
    # La liste à tester n'est pas modifiée
    
    # Copie de la liste à tester
    TestBis = np.copy(Test)
    
    for test in TestBis:
        prediction = knnPrediction(Entrainement,test,k)
        test[2] = prediction
    
    # Renvoi de la liste testée
    return TestBis

#######################################################################
# Comptage du nombre de données correctement catégorisées

def Comptage(Entrainement,Test,k):
    # Décompte des éléments bien catégorisés
    
    # On commence par catégoriser tous les éléments de la liste Test
    TestBis = knnTest(Entrainement,Test,k)
    
    compteur = 0
    for i in range(len(Test)):
        if Test[i][2]==TestBis[i][2]:
            compteur = compteur + 1
    
    return compteur

k = int(np.sqrt(NbDeDonnees))

print('Nombre   total   de   données    :',len(Donnees))
print('Nombre de données d\'entraînement :',len(DonneesEntrainement))
print('Nombre  de  données  à  tester   :',len(DonneesTest))
print('Nombre     de      classes       :',NbDeClasses)
print('k =',k)
print('delta =',delta)
TauxReussite = Comptage(DonneesEntrainement,DonneesTest,k) * 100 / len(DonneesTest)
print('Taux de réussite :',TauxReussite)
